﻿using System;
using System.Threading;
using System.Collections.Generic;

namespace Chapter01
{
    class Program
    {
        static void Main(string[] args)
        {
            // A static method
            StaticMethod();

            // Value types
            ValueTypes1();
            ValueTypes2();

            // Interfaces
            Interfaces();

            // Delegates
            Delegates1();

            // Generics
            Generics1();
            Generics2();
        }

        private static void StaticMethod()
        {
            int five = Calculator.Add(2, 3);
            Console.WriteLine(five);
        }

        private static void ValueTypes1()
        {
            Point pt = new Point(1, 2, 3);
            Console.WriteLine(pt.X);
            Console.WriteLine(pt.Y);
            Console.WriteLine(pt.Z);
        }

        private static void ValueTypes2()
        {
            Color red = Color.Red;
            Console.WriteLine(red);
        }

        private static void Interfaces()
        {
            IEnumerable collection = new MyCollection();
            IEnumerator enumerator = collection.GetEnumerator();
            if (enumerator.MoveNext())
            {
                Console.WriteLine(enumerator.Current);
            }
        }

        private static void Delegates1()
        {
            var t = new Thread(delegate(object data)
            {
                // Process customer object in the background
                Customer c = (Customer)data;
                Console.WriteLine(c.Name + " on " + Thread.CurrentThread.ManagedThreadId);
            });

            Customer customer = new Customer();
            customer.Name = "Bart";
            t.Start(customer);
        }

        private static void Delegates2()
        {
            new WorkingWithDelegates().Sample();
        }

        private static void Generics1()
        {
            var primes = new List<int> { 2, 3, 5, 7 };
            int firstPrime = primes[0];
            Console.WriteLine(firstPrime);
        }

        private static void Generics2()
        {
            try
            {
                System.Collections.ArrayList primes = new System.Collections.ArrayList();
                primes.Add(2);
                primes.Add(3);
                primes.Add(5);
                primes.Add(7);
                primes[0] = "Bart"; // Wouldn’t be valid on a List<int>
                int firstPrime = (int)primes[0]; // Invalid conversion causes runtime exception
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
            }
        }
    }

    #region Classes, Structures, and Enumerations

    static class Calculator
    {
        public static int Add(int a, int b)
        {
            return a + b;
        }
    }

    struct Point
    {
        private readonly int _x;
        private readonly int _y;
        private readonly int _z;

        public Point(int x, int y, int z)
        {
            _x = x; _y = y; _z = z;
        }

        public int X { get { return _x; } }
        public int Y { get { return _y; } }
        public int Z { get { return _z; } }
    }

    enum Color
    {
        Red,
        Green,
        Blue
    }

    #endregion

    #region Interfaces

    interface IEnumerable
    {
        IEnumerator GetEnumerator();
    }

    interface IEnumerator
    {
        bool MoveNext();
        object Current { get; }
        void Reset();
    }

    class MyCollection : IEnumerable
    {
        public IEnumerator GetEnumerator()
        {
            return new MyEnumerator();
        }

        class MyEnumerator : IEnumerator
        {
            public bool MoveNext()
            {
                return true;
            }

            public object Current
            {
                get { return "Hello"; }
            }

            public void Reset()
            {
                throw new NotImplementedException();
            }
        }
    }

    #endregion

    #region Delegates

    // Helper class for threading sample shown in Main method
    class Customer
    {
        public string Name
        {
            get;
            set;
        }
    }

    delegate int BinaryOperation(int a, int b);

    class WorkingWithDelegates
    {
        public int Calculate(BinaryOperation op, int a, int b)
        {
            // calling through the delegate
            return op(a, b);
        }

        public void Sample()
        {
            int three = Calculate(delegate(int a, int b) { return a + b; }, 1, 2);
            Console.WriteLine(three);
        }
    }

    #endregion
}
