﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Chapter05
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void EvaluationStack()
        {
            int a = 1;
            int b = 2;
            int c = 3;

            // Look in ILDASM how this works.
            int res = a + b * c;

            Console.WriteLine(res);
        }

        static void IntegerArithmetic()
        {
            int a = 6;
            int b = 7;
            int mul = a * b; // 42
            int add = mul + a; // 48
            int div = add / b; // 6
            int sub = mul - div; // 36
            int rem = sub % b; // 1

            Console.WriteLine(mul);
            Console.WriteLine(add);
            Console.WriteLine(div);
            Console.WriteLine(sub);
            Console.WriteLine(rem);
        }

        static void DecimalArithmetic()
        {
            decimal res = (1m / 3m) * 3m;

            Console.WriteLine(res);
        }

        static void CharacterArithmetic()
        {
            int numberOfLetters = 'z' - 'a' + 1;
            char e = (char)('a' + 4);

            Console.WriteLine(numberOfLetters);
            Console.WriteLine(e);
        }

        static void Overflow()
        {
            // Without checked keyword
            {
                int max = int.MaxValue;
                int res = max + 1;
                Console.WriteLine(res);
            }

            // With checked keyword - will throw an exception
            {
                int max = int.MaxValue;
                int res = checked(max + 1);
                Console.WriteLine(res);
            }

            // Block notation
            checked
            {
                int max = int.MaxValue;
                int res = max + 1;
                Console.WriteLine(res);
            }
        }

        static void StringConcatenation()
        {
            Console.Write("Enter your name: ");
            string name = Console.ReadLine();
            string message = "Hello " + name;
            Console.WriteLine(message);
        }

        static void StringConcatenation_Expensive()
        {
            string[] names = new[] { "Bart", "John", "Lisa" };

            string people = "";
            foreach (string name in names)
            {
                people += name + ", ";
            }

            // Look at number of intermediate string allocations in the Memory Window.
            Console.WriteLine(people);
        }

        static void StringConcatenation_Better()
        {
            string[] names = new[] { "Bart", "John", "Lisa" };

            var peopleList = new StringBuilder();
            foreach (string name in names)
            {
                peopleList.Append(name);
                peopleList.Append(", ");
            }

            string people = peopleList.ToString();
            Console.WriteLine(people);
        }

        static void Shifts()
        {
            int left = 3;
            int lres = left << 5; // 96
            int right = 42;
            int rres = right >> 3; // 5

            Console.WriteLine(lres);
            Console.WriteLine(rres);
        }

        static void LiftedOperators()
        {
            int? a = 5;
            int? b = null;
            bool gt = a > b;  // false
            bool le = a <= b; // false
            bool eq = a == b; // false
            bool ne = a != b; // true

            Console.WriteLine(gt);
            Console.WriteLine(le);
            Console.WriteLine(eq);
            Console.WriteLine(ne);
        }

        static void LogicalOperators()
        {
            int x = 5;
            int y = 3;
            int and = x & y; // 1
            int or = x | y;  // 7
            int xor = x ^ y; // 6
            int compX = ~x;  // -6

            Console.WriteLine(and);
            Console.WriteLine(or);
            Console.WriteLine(xor);
            Console.WriteLine(compX);
        }

        static void Conditional()
        {
            int x = -42;

            int absOfX = x >= 0 ? x : -x;
            Console.WriteLine(absOfX);
        }

        static void NullCoalescing()
        {
            var name = default(string); // null
            var res = name ?? "Person X";

            Console.WriteLine(res);
        }

        static void SimpleAssignment()
        {
            // Same as:  int x = 42;
            int x;
            x = 42;

            Console.WriteLine(x);
        }

        static void CompoundAssignment()
        {
            int a = 42;
            a += 6; // a = 48
            a /= 8; // a = 6
            Console.WriteLine(a);

            char c = 'z';
            c -= (char)2; // c = 'x'
            Console.WriteLine(c);

            DateTime d = DateTime.Now;
            d += TimeSpan.FromDays(5);
            Console.WriteLine(d);
        }

        static void DefiniteAssignment()
        {
            // This won't work:  int x;
            int x = 0; // Or, in case the original code had a bug, another value...
            Console.WriteLine(x);
        }

        static void PreAndPostIncrement()
        {
            // Pre
            {
                int x = 0;

                // The value after the increment is assigned, so y will be set to 1.
                int y = ++x;

                // At this point, x will have been incremented to 1.
                Console.WriteLine(x);
                Console.WriteLine(y); // 0
            }

            // Post
            {
                int x = 0;

                // The value before the increment is assigned, so y will be set to 0.
                int y = x++;

                // At this point, x will have been incremented to 1.
                Console.WriteLine(x);
                Console.WriteLine(y); // 1
            }
        }
    }
}
