﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Linq.Expressions;
using System.Diagnostics;
using System.Threading.Tasks;

namespace Chapter07
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void Selection_If()
        {
            var rand = new Random();
            int n = rand.Next(0, 100);
            PrintEvenOrNot(n);
        }

        static void PrintEvenOrNot(int n)
        {
            if (n % 2 == 0)
            {
                Console.WriteLine("Even");
            }
            else
            {
                Console.WriteLine("Odd");
            }
        }

        static void Selection_Switch()
        {
            PrintMessageForColor(Color.Red);
            PrintMessageForColor(Color.Green);
            PrintMessageForColor(Color.Blue);
        }

        static void PrintMessageForColor(Color color)
        {
            switch (color)
            {
                case Color.Red:
                    Console.WriteLine("Red roses.");
                    break;
                case Color.Green:
                    Console.WriteLine("Green grass.");
                    break;
                case Color.Blue:
                    Console.WriteLine("Blue skies.");
                    break;
                default:
                    throw new InvalidOperationException("Expected primary color.");
            }
        }

        static bool IsPrimary(Color color)
        {
            switch (color)
            {
                case Color.Red:
                case Color.Green:
                case Color.Blue:
                    return true;
                default:
                    return false;
            }
        }

        static void Selection_Switch_Eval()
        {
            var res = Eval(Expression.Add(Expression.Multiply(Expression.Constant(2), Expression.Constant(3)), Expression.Constant(1)));
            Console.WriteLine(res);
        }

        static int Eval(Expression ex)
        {
            switch (ex.NodeType)
            {
                case ExpressionType.Constant:
                    {
                        var ce = (ConstantExpression)ex;
                        if (!(ce.Value is int))
                            throw new InvalidOperationException("Only int supported.");
                        return (int)ce.Value;
                    }
                case ExpressionType.Add:
                    {
                        var be = (BinaryExpression)ex;
                        return Eval(be.Left) + Eval(be.Right);
                    }
                case ExpressionType.Multiply:
                    {
                        var be = (BinaryExpression)ex;
                        return Eval(be.Left) * Eval(be.Right);
                    }
                // Other binary operations.
                default:
                    throw new NotSupportedException("Can’t evaluate " + ex + ".");
            }
        }

        static int Eval2(Expression ex)
        {
            switch (ex.NodeType)
            {
                case ExpressionType.Constant:
                    return EvalConstant((ConstantExpression)ex);
                case ExpressionType.Add:
                    return EvalBinary((BinaryExpression)ex, (a, b) => a + b);
                case ExpressionType.Multiply:
                    return EvalBinary((BinaryExpression)ex, (a, b) => a * b);
                // Other binary operations.
                default:
                    throw new NotSupportedException("Can’t evaluate " + ex + ".");
            }
        }

        static int EvalConstant(ConstantExpression ce)
        {
            if (!(ce.Value is int))
                throw new InvalidOperationException("Only int supported.");
            return (int)ce.Value;
        }

        static int EvalBinary(BinaryExpression be, Func<int, int, int> f)
        {
            return f(Eval(be.Left), Eval(be.Right));
        }

        static string ToDutchBinaryString1(int x)
        {
            // Look in ILDASM
            switch (x)
            {
                case 0:
                    return "nul";
                case 1:
                    return "een";
                default:
                    throw new InvalidOperationException("Not a binary digit.");
            }
        }

        static string ToDutchBinaryString2(int x)
        {
            // Look in ILDASM
            switch (x)
            {
                case 11:
                    return "elf";
                case 13:
                    return "dertien";
                default:
                    throw new InvalidOperationException("Not a binary digit.");
            }
        }

        static string DutchDecimalDigitToEnglish(string s)
        {
            // Look in ILDASM
            switch (s)
            {
                case "nul": return "zero";
                case "een": return "one";
                case "twee": return "two";
                case "drie": return "three";
                case "vier": return "four";
                case "vijf": return "five";
                case "zes": return "six";
                case "zeven": return "seven";
                case "acht": return "eight";
                case "negen": return "nine";
                default:
                    throw new InvalidOperationException("Not a decimal digit.");
            }
        }

        static void While()
        {
            int i = 1;
            while (i <= 10)
            {
                Console.WriteLine(new string('*', i));
                i++;
            }
        }

        static void Break()
        {
            var rand = new Random();
            while (true)
            {
                var num = rand.Next();
                if (num % 2 == 0)
                    break;
                Console.WriteLine(num);
            }
        }

        static void For()
        {
            var primes = new[] { 2, 3, 5, 7 };
            // Forward loop over the array.
            for (int i = 0; i < primes.Length; i++)
            {
                Console.WriteLine(primes[i]);
            }
            // Backwards loop over the array.
            for (int i = primes.Length - 1; i >= 0; i--)
            {
                Console.WriteLine(primes[i]);
            }

            for (int exp = 0, pow = 1; exp <= 10; exp++, pow *= 2)
            {
                Console.WriteLine("2^{0} = {1}", exp, pow); // Using a string formatter.
            }
        }

        static void PrintPrimes(int max)
        {
            for (int num = 2; num < max; num++)
            {
                bool foundDiv = false;

                for (int div = 2; div <= Math.Sqrt(num); div++)
                {
                    if (num % div == 0)
                    {
                        foundDiv = true;
                        break;
                    }
                }

                if (!foundDiv)
                {
                    Console.WriteLine(num);
                }
            }
        }

        static void Foreach()
        {
            // Array
            {
                int[] primes = new int[] { 2, 3, 5, 7 };
                foreach (int prime in primes)
                {
                    Console.WriteLine(prime);
                }
            }

            // Other collection types
            {
                var primes = new List<int> { 2, 3, 5, 7 };
                foreach (int prime in primes)
                {
                    Console.WriteLine(prime);
                }

                var ages = new Dictionary<string, int> { { "Bart", 26 }, { "John", 62 } };
                foreach (string name in ages.Keys)
                {
                    Console.WriteLine("{0} is {1}", name, ages[name]);
                }
            }

            // LINQ
            {
                const int MB = 1024 * 1024;
                var memHungry = from proc in Process.GetProcesses()
                                where proc.WorkingSet64 > 50 * MB
                                select new { proc.ProcessName, MB = proc.WorkingSet64 / MB };

                foreach (var memoryHungryProcess in memHungry)
                {
                    Console.WriteLine(memoryHungryProcess);
                }
            }
        }

        static void Iterators_ManualPlumbing()
        {
            foreach (int prime in new PrimeGenerator())
            {
                Console.WriteLine(prime);
                Console.Write("Want More? ");
                if (!Console.ReadLine().ToLower().StartsWith("y"))
                    break;
            }
        }

        class PrimeGenerator
        {
            public PrimeEnumerator GetEnumerator()
            {
                return new PrimeEnumerator();
            }
        }

        class PrimeEnumerator
        {
            private int _current = 1;

            public bool MoveNext()
            {
                for (int num = _current + 1; num < int.MaxValue; num++)
                {
                    bool foundDiv = false;

                    for (int div = 2; div <= Math.Sqrt(num); div++)
                    {
                        if (num % div == 0)
                        {
                            foundDiv = true;
                            break;
                        }
                    }

                    if (!foundDiv)
                    {
                        _current = num;
                        return true;
                    }
                }

                return false;
            }

            public int Current
            {
                get { return _current; }
            }
        }

        static void Iterators_TheEasyWay()
        {
            foreach (int prime in GetPrimes(int.MaxValue))
            {
                Console.WriteLine(prime);
                Console.Write("Want More? ");
                if (!Console.ReadLine().ToLower().StartsWith("y"))
                    break;
            }
        }

        static IEnumerable<int> GetPrimes(int max)
        {
            for (int num = 2; num < max; num++)
            {
                bool foundDiv = false;

                for (int div = 2; div <= Math.Sqrt(num); div++)
                {
                    if (num % div == 0)
                    {
                        foundDiv = true;
                        break;
                    }
                }

                if (!foundDiv)
                {
                    yield return num;
                }
            }
        }

        static void ParallelLoops()
        {
            // Don't expect results to come out in order...
            Parallel.For(1, 100, i =>
            {
                Console.WriteLine(i);
            });
        }

        static void PLINQ()
        {
            const int MB = 1024 * 1024;

            var memHungry = from proc in Process.GetProcesses().AsParallel()
                            where proc.WorkingSet64 > 50 * MB
                            select new { proc.ProcessName, MB = proc.WorkingSet64 / MB };

            foreach (var hungry in memHungry)
                Console.WriteLine(hungry);
        }

        static void Return()
        {
            PrintName("Bart");
        }

        static void PrintName(string name)
        {
            if (name == null)
                return;
            Console.WriteLine(name);
        }
    }

    enum Color
    {
        Red,
        Green,
        Blue,
    }
}