﻿using System;
using System.Collections;
using System.Collections.Generic;

namespace Chapter09
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void Classes()
        {
            Location loc1 = new Location();
            Location loc2 = loc1;

            loc1.X = 5;
            loc1.Y = 3;

            Console.WriteLine(loc1.X + "," + loc1.Y); // 5,3
            Console.WriteLine(loc2.X + "," + loc2.Y); // 5,3 due to aliasing (reference)
        }

        static void Structs()
        {
            Point loc1 = new Point();
            Point loc2 = loc1;

            loc1.X = 5;
            loc1.Y = 3;

            Console.WriteLine(loc1.X + "," + loc1.Y); // 5,3
            Console.WriteLine(loc2.X + "," + loc2.Y); // 0,0
        }

        static void Boxing()
        {
            Bar(42); // Causes boxing
            Foo(42); // No boxing
        }

        static void Bar(object value)
        {
            // For some mysterious reason, we know we’ll always
            // get an Int32 through the value parameter.
            int a = (int)value;
            Console.WriteLine(a);
        }

        static void Foo(int value)
        {
            Console.WriteLine(value);
        }

        static void Boxing_NonGenericCollections()
        {
            var lst = new ArrayList();
            for (int i = 0; i < 1000000; i++)
                lst.Add(i); // 1000000 boxes, expected parameter statically typed as object.

            var sum = 0L;
            foreach (int i in lst)
                sum += i;

            Console.WriteLine(sum);
        }

        static void NoBoxing_GenericCollections()
        {
            var lst = new List<int>();
            for (int i = 0; i < 1000000; i++)
                lst.Add(i); // No boxes needed, expected parameter statically typed as int.

            var sum = 0L;
            foreach (int i in lst)
                sum += i;

            Console.WriteLine(sum);
        }

        static void DangerOfMutability()
        {
            var kid = new Kid();
            Console.WriteLine(kid.Location.X + ", " + kid.Location.Y);

            // The following two lines don't compile (luckily).
            //kid.Location.X = 3;
            //kid.Location.Y = 4;
            //Console.WriteLine(kid.Location.X + ", " + kid.Location.Y);
        }

        static void StaticVersusInstance()
        {
            // Singleton retrieval through static property
            var svc = PaymentService.Service;

            // Instance method invocation
            svc.SchedulePayment(42, 24, 123.45m);

            // Making a new instance
            var mrx = new Person();

            // Population is a static property shared for the whole population of People instances
            Console.WriteLine(Person.Population);
        }
    }

    class Kid
    {
        public Point Location { get; set; }
    }
    
    struct Point
    {
        public int X { get; set; }
        public int Y { get; set; }
    }

    class Location
    {
        public int X { get; set; }
        public int Y { get; set; }
    }

    // Singleton implementation
    sealed class PaymentService
    {
        private static PaymentService s_service;
        private Queue<Payment> _payments;

        private PaymentService()
        {
            // This is a private constructor; therefore it can only
            // be called from inside the PaymentService type.
            _payments = new Queue<Payment>();
        }

        public static PaymentService Service
        {
            get
            {
                // Create service the first time it’s requested. For
                // the remainder of the program’s lifetime we’ll be
                // using this singleton instance.
                if (s_service == null)
                {
                    s_service = new PaymentService();
                }
                return s_service;
            }
        }

        public void SchedulePayment(int fromAccount, int toAccount, decimal amount)
        {
            _payments.Enqueue(new Payment(fromAccount, toAccount, amount));
        }
    }

    class Payment
    {
        private int fromAccount;
        private int toAccount;
        private decimal amount;

        public Payment(int fromAccount, int toAccount, decimal amount)
        {
            this.fromAccount = fromAccount;
            this.toAccount = toAccount;
            this.amount = amount;
        }
    }

    class Person
    {
        private static int s_population;

        public Person()
        {
            s_population++;
        }

        public static int Population
        {
            get { return s_population; }
        }
    }
}

// Illustration of some visibility concepts
namespace HR
{
    class Person
    {
        public string Name { get; set; }
        public int Age { get; set; }

        class Jewelry // Only accessible to code inside the Person class
        {
        }
    }
}