﻿using System;
using System.Collections;
using System.Collections.Generic;

namespace Chapter16
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void NonGeneric_ArrayList()
        {
            var lst = new ArrayList();
            lst.Add(1);                               // { 1 }
            lst.AddRange(new[] { 2, 3, 7, 6, 5, 4 }); // { 1, 2, 3, 7, 6, 5, 4 }
            lst.Insert(index: 0, value: 0);           // { 0, 1, 2, 3, 7, 6, 5, 4 }
            lst.RemoveAt(index: 4);                   // { 0, 1, 2, 3, 6, 5, 4 }
            lst.Remove(obj: 1);                       // { 0, 2, 3, 6, 5, 4 }
            lst.RemoveRange(index: 2, count: 3);      // { 0, 2, 4 }
            lst.Reverse();                            // { 4, 2, 0 }
            lst.Sort();                               // { 0, 2, 4 }
            bool hasTwo = lst.Contains(2);            // true
            int two = (int)lst[1];                    // 2
        }

        static void NonGeneric_Hashtable()
        {
            var tbl = new Hashtable();
            tbl.Add("Bart", 26);
            tbl["John"] = 62;

            bool bartIsKnown = tbl.ContainsKey("Bart");
            bool someIsAged62 = tbl.ContainsValue(62);
            foreach (string name in tbl.Keys)
                Console.WriteLine("Age of {0} is {1}", name, tbl[name]);

            tbl.Remove("Bart");
        }

        static void NonGeneric_Queue()
        {
            var postOffice = new Queue();
            postOffice.Enqueue(new Person("Bart", 10));
            postOffice.Enqueue(new Person("Homer", 38));
            postOffice.Enqueue(new Person("Ned", 60));

            var bart = (Person)postOffice.Dequeue();
            var homer = (Person)postOffice.Peek(); // D’oh
            int twoLeft = postOffice.Count;
        }

        static void NonGeneric_Stack()
        {
            var stack = new Stack();
            stack.Push(1);
            stack.Push(2);
            stack.Push(3);
            var three = (int)stack.Pop();
            var two = (int)stack.Peek();
            var twoLeft = stack.Count;
        }

        static void NonGeneric_BitArray()
        {
            const int N = 20;

            var first = new BitArray(N, false);
            for (int i = 0; i < first.Length; i += 2)
                first[i] = true;

            var second = new BitArray(N, false);
            for (int i = 1; i < second.Length; i += 2)
                second[i] = true;

            Print("First: ", first);           // 10101010101010101010
            Print("Second: ", second);         // 01010101010101010101
            Print("Xor: ", first.Xor(second)); // 11111111111111111111
        }

        static void Print(string title, BitArray bitArray)
        {
            Console.Write(title);
            foreach (bool b in bitArray)
                Console.Write(b ? "1" : "0");

            Console.WriteLine();
        }

        static void Generic_List()
        {
            var lst = new List<int>();
            lst.Add(1);                               // { 1 }
            lst.AddRange(new[] { 2, 3, 7, 6, 5, 4 }); // { 1, 2, 3, 7, 6, 5, 4 }
            lst.Insert(index: 0, item: 0);            // { 0, 1, 2, 3, 7, 6, 5, 4 }
            lst.RemoveAt(index: 4);                   // { 0, 1, 2, 3, 6, 5, 4 }
            lst.Remove(item: 1);                      // { 0, 2, 3, 6, 5, 4 }
            lst.RemoveRange(index: 2, count: 3);      // { 0, 2, 4 }
            lst.Reverse();                            // { 4, 2, 0 }
            lst.Sort();                               // { 0, 2, 4 }
            bool hasTwo = lst.Contains(2);            // true
            int two = lst[1];                         // 2
        }

        static void Generic_List_Methods()
        {
            var lst = new List<int> { 1, 2, 3, 4, 5, 6, 7 };

            // Convert all elements and create a new list.
            List<string> numbers = lst.ConvertAll(n => n.ToString());

            // Condition holds for any element (Exists) or all elements (TrueForAll)?
            bool hasLargerThanFive = lst.Exists(n => n > 5); // true
            bool noElementIsThirtheen = lst.TrueForAll(n => n != 13); // true

            // Find element(s) or indexes.
            int firstEven = lst.Find(n => n % 2 == 0); // 2
            int lastEven = lst.FindLast(n => n % 2 == 0); // 6
            List<int> allOdds = lst.FindAll(n => n % 2 != 0); // { 1, 3, 5, 7 }
            int indexOfFirstLargerThanThree = lst.FindIndex(n => n > 3); // 3
            int indexOfLastLessThanFive = lst.FindLastIndex(n => n < 5); // 3
        }

        static void Generic_LinkedList()
        {
            var lst = new LinkedList<string>();

            var elem1 = lst.AddFirst("isn’t");
            var elem2 = lst.AddBefore(elem1, "This");
            var elem3 = lst.AddAfter(elem1, "a");
            var elem4 = lst.AddLast("list");
            var elem5 = lst.AddBefore(elem4, "singly");
            var elem6 = lst.AddAfter(elem5, "linked");

            lst.Last.Previous.Previous.Value = "doubly";
            lst.First.Next.Value = "is";

            foreach (var word in lst)
                Console.Write(word + " ");

            Console.WriteLine();
        }

        static void Generic_Dictionary()
        {
            var tbl = new Dictionary<string, int>();
            tbl.Add("Bart", 26);
            tbl["John"] = 62;

            bool bartIsKnown = tbl.ContainsKey("Bart");
            bool someIsAged62 = tbl.ContainsValue(62);
            foreach (var name in tbl.Keys)
                Console.WriteLine("Age of {0} is {1}", name, tbl[name]);

            tbl.Remove("Bart");

            int ageOfMarge;
            if (tbl.TryGetValue("Marge", out ageOfMarge))
            {
                // Code if the key was found, with ageOfMarge set to the retrieved value.
            }
        }

        static void Generic_Sorted()
        {
            var ages = new SortedDictionary<string, int>(new ReverseComparer<string>()) {
                { "Bart", 10 },
                { "Marge", 34 },
                { "Homer", 38 }
            };

            foreach (var name in ages.Keys)
                Console.WriteLine("Age of {0} is {1}", name, ages[name]);
        }

        class ReverseComparer<T> : IComparer<T>
        {
            public int Compare(T x, T y)
            {
                return -Comparer<T>.Default.Compare(x, y);
            }
        }

        static void Generic_HashSet()
        {
            var people = new HashSet<Person>();
            people.Add(new Person("Bart", 10));
            people.Add(new Person("Homer", 38));
            people.Add(new Person("Lisa", 8));
            people.Add(new Person("Marge", 34));
            people.Add(new Person("Lisa", 8)); // Duplicate will not be added.

            int four = people.Count;
        }

        static void Generic_HashSet_Operations()
        {
            var numbers = new HashSet<int> { 1, 2, 2, 3, 4, 5, 5 }; // { 1, 2, 3, 4, 5 }
            numbers.Add(6);                                         // { 1, 2, 3, 4, 5, 6 }
            numbers.Remove(3);                                      // { 1, 2, 4, 5, 6 }
            numbers.RemoveWhere(n => n % 2 == 0);                   // { 1, 5 }
            numbers.Clear();                                        // { }

            foreach (int number in numbers)
                Console.WriteLine(number);
        }

        static void Generic_SortedSet()
        {
            var sortedNumbers = new SortedSet<int> { 5, 2, 7, 11, 3 };

            // Prints 2, 3, 5, 7, 11
            foreach (int number in sortedNumbers)
                Console.WriteLine(number);
        }

        static void Generic_Queue()
        {
            var postOffice = new Queue<Person>();
            postOffice.Enqueue(new Person("Bart", 10));
            postOffice.Enqueue(new Person("Homer", 38));
            postOffice.Enqueue(new Person("Ned", 60));

            var bart = postOffice.Dequeue();
            var homer = postOffice.Peek(); // D’oh
            int twoLeft = postOffice.Count;
        }

        static void Generic_Stack()
        {
            var stack = new Stack<int>();
            stack.Push(1);
            stack.Push(2);
            stack.Push(3);

            var three = stack.Pop();
            var two = stack.Peek();
            var twoLeft = stack.Count;
        }
    }

    class Person
    {
        private readonly string _name;
        private readonly int _age;

        public Person(string name, int age)
        {
            _name = name;
            _age = age;
        }

        public string Name
        {
            get { return _name; }
        }

        public int Age
        {
            get { return _age; }
        }

        public override int GetHashCode()
        {
            return (_name != null ? _name.GetHashCode() : 0) ^ _age;
        }
    }
}
