﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Data.SqlClient;
using System.Diagnostics;
using System.Linq;
using System.Xml;
using System.Xml.Linq;

namespace Chapter19
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void WithoutLinq_Memory()
        {
            var products = new List<Product> {
                new Product { Name = "Chai", Price = 18.00m },
                new Product { Name = "Chang", Price = 19.00m },
                new Product { Name = "Aniseed Syrup", Price = 10.00m },
                new Product { Name = "Chef Anton’s Cajun Seasoning", Price = 22.00m },
                new Product { Name = "Chef Anton’s Gumbo Mix", Price = 21.35m },
                new Product { Name = "Grandma’s Boysenberry Spread", Price = 25.00m },
                new Product { Name = "Uncle Bob’s Organic Dried Pears", Price = 30.00m },
                new Product { Name = "Northwoods Cranberry Sauce", Price = 40.00m },
                new Product { Name = "Mishi Kobe Niku", Price = 97.00m },
                new Product { Name = "Ikura", Price = 31.00m }
            };

            // Filtering
            var cheap = new List<Product>();
            foreach (var product in products)
            {
                if (product.Price < 25)
                    cheap.Add(product);
            }

            // Grouping
            var priceGroups = new Dictionary<int, List<Product>>();
            foreach (var product in products)
            {
                int group = 10 * (int)(product.Price / 10);
                List<Product> productsInGroup;
                if (!priceGroups.TryGetValue(group, out productsInGroup))
                    priceGroups[group] = productsInGroup = new List<Product>();
                productsInGroup.Add(product);
            }
        }

        static void WithoutLinq_Relational()
        {
            // Refer to the documentation for SqlConnection to supply a valid connection string.
            var connectionString = "<enter a connection string here>";

            var products = new List<Product>();
            using (var conn = new SqlConnection(connectionString))
            {
                //
                // *** Safer ***
                //
                // string sql = "SELECT * FROM Products WHERE UnitPrice < @UnitPrice";
                // 
                string sql = "SELECT * FROM Products WHERE UnitPrice < 25";

                using (var cmd = new SqlCommand(sql, conn))
                {
                    //
                    // *** Safer ***
                    //
                    // var maxPrice = cmd.Parameters.Add("@Price", SqlDbType.Decimal);
                    // maxPrice.Value = 25m; /* get in from somewhere, e.g. user input */
                    //

                    conn.Open();
                    using (var reader = cmd.ExecuteReader())
                    {
                        while (reader.Read())
                        {
                            string name = (string)reader["ProductName"];
                            decimal price = (decimal)reader["UnitPrice"];
                            products.Add(new Product { Name = name, Price = price });
                        }
                    }
                }
            }
        }

        static void WithoutLinq_Xml()
        {
            /*
             * Save this piece of XML in c:\temp\products.xml
             * 
            <Products>
                <Product Name="Chai" Price="18.00" />
                <Product Name="Chang" Price="19.00" />
                <Product Name="Aniseed Syrup" Price="10.00" />
                <Product Name="Chef Anton’s Cajun Seasoning" Price="22.00" />
                <Product Name="Chef Anton’s Gumbo Mix" Price="21.35" />
                <Product Name="Grandma’s Boysenberry Spread" Price="25.00" />
                <Product Name="Uncle Bob’s Organic Dried Pears" Price="30.00" />
                <Product Name="Northwoods Cranberry Sauce" Price="40.00" />
                <Product Name="Mishi Kobe Niku" Price="97.00" />
                <Product Name="Ikura" Price="31.00" />
            </Products>
            */

            XmlDocument doc = new XmlDocument();
            doc.Load(@"c:\temp\products.xml");
            XmlNodeList res = doc.SelectNodes("//Product[@Price<25]");
            var products = new List<Product>();
            foreach (XmlNode node in res)
            {
                products.Add(new Product
                {
                    Name = node.Attributes["Name"].Value,
                    Price = decimal.Parse(node.Attributes["Price"].Value)
                });
            }
        }

        static void WithLinq_Memory()
        {
            var products = new List<Product> {
                new Product { Name = "Chai", Price = 18.00m },
                new Product { Name = "Chang", Price = 19.00m },
                new Product { Name = "Aniseed Syrup", Price = 10.00m },
                new Product { Name = "Chef Anton’s Cajun Seasoning", Price = 22.00m },
                new Product { Name = "Chef Anton’s Gumbo Mix", Price = 21.35m },
                new Product { Name = "Grandma’s Boysenberry Spread", Price = 25.00m },
                new Product { Name = "Uncle Bob’s Organic Dried Pears", Price = 30.00m },
                new Product { Name = "Northwoods Cranberry Sauce", Price = 40.00m },
                new Product { Name = "Mishi Kobe Niku", Price = 97.00m },
                new Product { Name = "Ikura", Price = 31.00m }
            };

            // Filtering
            decimal upperPrice = 25m;
            var cheap = from product in products
                        where product.Price < upperPrice
                        select product;

            foreach (var product in cheap)
                Console.WriteLine(product.Name);

            // Restriction
            var fiveCheap = (from product in products
                             where product.Price < upperPrice
                             select product).Take(5);

            foreach (var product in fiveCheap)
                Console.WriteLine(product.Name);

            // Grouping
            var priceGroups = from product in products
                              group product by 10m * (int)(product.Price / 10);

            foreach (var priceGroup in priceGroups)
            {
                Console.WriteLine("From {0} to {1}", priceGroup.Key, priceGroup.Key + 9.99m);

                foreach (var product in priceGroup)
                    Console.WriteLine(" {0} costs ${1}", product.Name, product.Price);
            }
        }

        static void WithLinq_Relational()
        {
            //
            // Please follow the instructions from the book to create a LINQ to SQL data context.
            // Then run the code below.
            //

            /*
            using (var ctx = new NorthwindDataContext())
            {
                var cheap = from product in ctx.Products
                            where product.UnitPrice < 25
                            select product;
             
                foreach (var product in cheap)
                    Console.WriteLine("{0} costs {1}", product.ProductName,
                    product.UnitPrice);
            }
             */
        }

        static void WithLinq_Xml()
        {
            /*
             * Save this piece of XML in c:\temp\products.xml
             * 
            <Products>
                <Product Name="Chai" Price="18.00" />
                <Product Name="Chang" Price="19.00" />
                <Product Name="Aniseed Syrup" Price="10.00" />
                <Product Name="Chef Anton’s Cajun Seasoning" Price="22.00" />
                <Product Name="Chef Anton’s Gumbo Mix" Price="21.35" />
                <Product Name="Grandma’s Boysenberry Spread" Price="25.00" />
                <Product Name="Uncle Bob’s Organic Dried Pears" Price="30.00" />
                <Product Name="Northwoods Cranberry Sauce" Price="40.00" />
                <Product Name="Mishi Kobe Niku" Price="97.00" />
                <Product Name="Ikura" Price="31.00" />
            </Products>
            */

            var doc = XDocument.Load(@"c:\temp\Products.xml");
            var products = doc.Element("Products").Elements();

            var cheap = from product in products
                        let name = product.Attribute("Name").Value
                        let price = decimal.Parse(product.Attribute("Price").Value)
                        where price < 25m
                        select new Product { Name = name, Price = price };

            foreach (var product in cheap)
                Console.WriteLine(product.Name + " costs " + product.Price);
        }

        static void FromClause()
        {
            var nums = new List<int> { 1, 2, 3, 4, 5 };

            // Hover over n to see the inferred type.
            var res = from n in nums
                      select n;

            foreach (var x in res)
                Console.WriteLine(x);
        }

        static void RangeIdentifierTypes()
        {
            var nums = new ArrayList { 1, 2, 3, 4, 5 };

            // Uses a call to Cast<int> underneath.
            var res = from int n in nums
                      select n;

            foreach (var x in res)
                Console.WriteLine(x);
        }

        static void SelectClause()
        {
            var nums = new List<int> { 1, 2, 3, 4, 5 };

            // Uses a call to Select.
            var twice = from n in nums
                        select n * 2;

            foreach (var x in twice)
                Console.WriteLine(x);
        }

        static void SelectClause_More()
        {
            var products = new List<Product> {
                new Product { Name = "Chai", Price = 18.00m },
                new Product { Name = "Chang", Price = 19.00m },
                new Product { Name = "Aniseed Syrup", Price = 10.00m },
                new Product { Name = "Chef Anton’s Cajun Seasoning", Price = 22.00m },
                new Product { Name = "Chef Anton’s Gumbo Mix", Price = 21.35m },
                new Product { Name = "Grandma’s Boysenberry Spread", Price = 25.00m },
                new Product { Name = "Uncle Bob’s Organic Dried Pears", Price = 30.00m },
                new Product { Name = "Northwoods Cranberry Sauce", Price = 40.00m },
                new Product { Name = "Mishi Kobe Niku", Price = 97.00m },
                new Product { Name = "Ikura", Price = 31.00m }
            };

            // Not incredibly useful but constants are expressions.
            var oneForEveryProduct = from product in products
                                     select 1;

            foreach (var one in oneForEveryProduct)
                Console.WriteLine(one);

            // Using an anonymous type.
            var nameAndPrice = from product in products
                               select new { product.Name, product.Price };

            foreach (var p in nameAndPrice)
                Console.WriteLine(p);

            // Using a closure over an outer variable (refer to chapter on delegates).
            decimal percent = 0.05m;
            var productDiscounts = from p in products
                                   select new { p.Name, Discount = p.Price * percent };

            foreach (var discounted in productDiscounts)
                Console.WriteLine(discounted);

            // Calling a method.
            var pricingLabels = from p in products
                                select string.Format("{0} costs ${1}", p.Name, p.Price);

            foreach (var label in pricingLabels)
                Console.WriteLine(label);
        }

        static void WhereClause()
        {
            var nums = new List<int> { 1, 2, 3, 4, 5 };

            // Uses a call to Where.
            var evens = from n in nums
                        where n % 2 == 0
                        select n;

            foreach (var x in evens)
                Console.WriteLine(x);
        }

        static void OrderByClause()
        {
            var nums = new List<int> { 3, 5, 4, 2, 1 };

            // Uses a call to OrderBy.
            var sorted = from n in nums
                         orderby n /* ascending */
                         select n;

            foreach (var x in sorted)
                Console.WriteLine(x);

            // Uses a call to OrderByDescending.
            var sortedDesc = from n in nums
                             orderby n descending
                             select n;

            foreach (var x in sortedDesc)
                Console.WriteLine(x);
        }

        static void OrderByClause_More()
        {
            var products = new List<Product> {
                new Product { Name = "Chai", Price = 18.00m },
                new Product { Name = "Chang", Price = 19.00m },
                new Product { Name = "Aniseed Syrup", Price = 10.00m },
                new Product { Name = "Chef Anton’s Cajun Seasoning", Price = 22.00m },
                new Product { Name = "Chef Anton’s Gumbo Mix", Price = 21.35m },
                new Product { Name = "Grandma’s Boysenberry Spread", Price = 25.00m },
                new Product { Name = "Uncle Bob’s Organic Dried Pears", Price = 30.00m },
                new Product { Name = "Northwoods Cranberry Sauce", Price = 40.00m },
                new Product { Name = "Mishi Kobe Niku", Price = 97.00m },
                new Product { Name = "Ikura", Price = 31.00m }
            };

            // Not incredibly useful but constants are expressions. Notice this doesn’t
            // guarantee to return the products in the input order!
            var orderByConstant = from product in products
                                  orderby 1
                                  select product;

            foreach (var p in orderByConstant)
                Console.WriteLine(p.Name);

            // Expensive to cheap requires descending sort.
            var decreasingPrice = from product in products
                                  orderby product.Price descending
                                  select product;

            foreach (var p in decreasingPrice)
                Console.WriteLine(p.Name + " costs " + p.Price);

            var priceThenName = from product in products
                                orderby product.Price, product.Name
                                select product;

            foreach (var p in priceThenName)
                Console.WriteLine(p.Name + " costs " + p.Price);
        }

        static void GroupByClause()
        {
            var byMod3 = from n in new[] { 3, 7, 1, 5, 0, 8, 0, 4, 2 }
                         group n by n % 3;

            foreach (var remainderGroup in byMod3)
            {
                Console.WriteLine("Remainder = " + remainderGroup.Key);
                foreach (var number in remainderGroup)
                {
                    Console.WriteLine(" " + number);
                }
            }
        }

        static void GroupByClause_More()
        {
            var products = new List<Product> {
                new Product { Name = "Chai", Price = 18.00m },
                new Product { Name = "Chang", Price = 19.00m },
                new Product { Name = "Aniseed Syrup", Price = 10.00m },
                new Product { Name = "Chef Anton’s Cajun Seasoning", Price = 22.00m },
                new Product { Name = "Chef Anton’s Gumbo Mix", Price = 21.35m },
                new Product { Name = "Grandma’s Boysenberry Spread", Price = 25.00m },
                new Product { Name = "Uncle Bob’s Organic Dried Pears", Price = 30.00m },
                new Product { Name = "Northwoods Cranberry Sauce", Price = 40.00m },
                new Product { Name = "Mishi Kobe Niku", Price = 97.00m },
                new Product { Name = "Ikura", Price = 31.00m }
            };

            var byFirstLetterOfName = from product in products
                                      group product by product.Name[0];

            foreach (var g in byFirstLetterOfName)
            {
                Console.WriteLine(g.Key);

                foreach (var p in g)
                    Console.WriteLine("  " + p.Name + " costs " + p.Price);
            }

            /* Use a LINQ to SQL data context for Northwind to run the following sample.
             * (Or add a Category and Supplier property to our Product class.)
             * 
            var byCategoryAndSupplier = from product in products
                                        group product by new
                                        {
                                            product.Category,
                                            product.Supplier
                                        };
             */

            var personnel = new[] {
                new { ID = 1, First = "Bart", Age = 26 },
                new { ID = 2, First = "John", Age = 62 },
                new { ID = 3, First = "Mike", Age = 24 },
                new { ID = 4, First = "John", Age = 53 },
                new { ID = 5, First = "Bart", Age = 26 },
                new { ID = 6, First = "Eric", Age = 26 },
                new { ID = 7, First = "John", Age = 62 },
                new { ID = 8, First = "Eric", Age = 30 },
                new { ID = 9, First = "Eric", Age = 26 },
            };

            var byNameAndAge = from employee in personnel
                               group employee by new
                               {
                                   employee.First,
                                   employee.Age
                               };

            foreach (var group in byNameAndAge)
            {
                Console.Write(group.Key + ": ");

                foreach (var employee in group)
                    Console.Write(employee.ID + " ");
                Console.WriteLine();
            }
        }

        static void JoinClause()
        {
            var processDescriptions = new[] {
                new { Name = "devenv", Description = "Microsoft Visual Studio" },
                new { Name = "dwm", Description = "Desktop Window Manager" },
                new { Name = "iexplore", Description = "Windows Internet Explorer" },
                new { Name = "sqlservr", Description = "Microsoft SQL Server" },
                new { Name = "svchost", Description = "Service Host" },
                new { Name = "WINWORD", Description = "Microsoft Office Word" }
            };

            const int MB = 1024 * 1024;
            var heavyMemoryUsers = (from p in Process.GetProcesses()
                                    join d in processDescriptions on p.ProcessName equals d.Name
                                    where p.WorkingSet64 > 100 * MB
                                    orderby p.WorkingSet64 descending
                                    select new
                                    {
                                        Name = d.Description,
                                        Memory = p.WorkingSet64 / MB
                                    })
                                   .Take(10);

            foreach (var p in heavyMemoryUsers)
                Console.WriteLine(p);
        }

        static void ContinuationOfQueryExpression_AfterSelect()
        {
            var products = new List<Product2> {
                new Product2 { ProductName = "Chai", UnitPrice = 18.00m },
                new Product2 { ProductName = "Chang", UnitPrice = 19.00m },
                new Product2 { ProductName = "Aniseed Syrup", UnitPrice = 10.00m },
                new Product2 { ProductName = "Chef Anton’s Cajun Seasoning", UnitPrice = 22.00m },
                new Product2 { ProductName = "Chef Anton’s Gumbo Mix", UnitPrice = 21.35m },
                new Product2 { ProductName = "Grandma’s Boysenberry Spread", UnitPrice = 25.00m },
                new Product2 { ProductName = "Uncle Bob’s Organic Dried Pears", UnitPrice = 30.00m },
                new Product2 { ProductName = "Northwoods Cranberry Sauce", UnitPrice = 40.00m },
                new Product2 { ProductName = "Mishi Kobe Niku", UnitPrice = 97.00m },
                new Product2 { ProductName = "Ikura", UnitPrice = 31.00m }
            };

            var expensiveProducts = from product in products
                                    select new
                                    {
                                        Name = product.ProductName,
                                        Price = product.UnitPrice
                                    } into product
                                    where product.Price > 100m
                                    select product;

            foreach (var p in expensiveProducts)
                Console.WriteLine(p.Name + " costs " + p.Price);
        }

        static void ContinuationOfQueryExpressions_AfterGroupBy()
        {
            var queryOperators = from m in typeof(Enumerable).GetMethods()
                                 group m by m.Name into queryOperator
                                 select new
                                 {
                                     Name = queryOperator.Key,
                                     Overloads = queryOperator.Count()
                                 };

            foreach (var queryOperator in queryOperators)
                Console.WriteLine("{0} has {1} overload(s)", queryOperator.Name, queryOperator.Overloads);
        }

        static void LetClause()
        {
            var res = from i in Enumerable.Range(0, 10)
                      let square = i * i
                      where square % 2 == 0
                      let cube = i * square
                      let stars = new string('*', i)
                      select new { Stars = stars, Cube = cube };

            foreach (var x in res)
                Console.WriteLine(x);
        }
    }

    class Product
    {
        public string Name { get; set; }
        public decimal Price { get; set; }
    }

    class Product2
    {
        public string ProductName { get; set; }
        public decimal UnitPrice { get; set; }
    }
}
