﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Extensibility;
using System.IO;
using System.Reflection;

namespace ExtensiCalc
{
    class Program
    {
        static void Main()
        {
            Console.WriteLine("ExtensiCalc - The world’s best calculator");
            Console.WriteLine();
            var operations = GetOperations().ToList();

            while (true)
            {
                int i = 1;
                foreach (var op in operations)
                {
                    var metadata = op.GetType().GetCustomAttributes(false).OfType<CalculationOperationAttribute>().SingleOrDefault();
                    Console.WriteLine(i++ + ". " + metadata.Name);
                }
                Console.WriteLine();

                int selectedOperation = -1;
                while (selectedOperation < 0 || selectedOperation > operations.Count)
                {
                    selectedOperation = ReadInt("Enter selection: ");
                }
                Console.WriteLine();

                if (selectedOperation == 0)
                    break;

                ICalculation operation = operations[selectedOperation - 1];
                var opa = operation.GetType().GetCustomAttributes(false).OfType<CalculationOperationAttribute>().SingleOrDefault();
                int a = ReadInt("a: ");
                int b = ReadInt("b: ");
                //Console.WriteLine("Result = " + operation.Calculate(a, b));
                Console.WriteLine("{0} {1} {2} = {3}", a, opa.Infix, b, operation.Calculate(a, b));
                Console.WriteLine();
                Console.WriteLine();
            }
        }

        static int ReadInt(string prompt)
        {
            int result = 0;
            bool valid = false;
            while (!valid)
            {
                Console.Write(prompt);
                valid = int.TryParse(Console.ReadLine(), out result);
            }
            return result;
        }

        static IEnumerable<ICalculation> GetOperations()
        {
            return GetBuiltinOperations().Concat(GetExtensions());
        }

        static IEnumerable<ICalculation> GetBuiltinOperations()
        {
            yield return new Addition();
            yield return new Subtraction();
            yield return new Multiplication();
            yield return new Division();
            // You can add other built-in operations here easily...
        }

        static IEnumerable<ICalculation> GetExtensions()
        {
            // Refer to the book to lay out the directory structure correctly.

            if (Directory.Exists("Extensions"))
            {
                var icalc = typeof(ICalculation);
                foreach (var file in Directory.EnumerateFiles("Extensions", "*.dll"))
                {
                    // For isolation, one will like to use AppDomains...
                    var asm = Assembly.LoadFrom(file);

                    // “LINQ to Reflection” to the rescue.
                    var extTypes = from t in asm.GetTypes()
                                   where t.GetInterfaces().Contains(icalc)
                                   select t;

                    foreach (var ext in extTypes)
                        yield return (ICalculation)Activator.CreateInstance(ext);
                }
            }
        }
    }
}
