﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;

using Word = Microsoft.Office.Interop.Word;
using Excel = Microsoft.Office.Interop.Excel;

namespace Chapter22_OfficeInterop_NET35
{
    class Program
    {
        static void Main(string[] args)
        {
            var excel = new Excel.Application();
            excel.Visible = true;

            // New empty workbook
            excel.Workbooks.Add(Type.Missing);

            // Column headers
            Excel.Worksheet sheet = (Excel.Worksheet)excel.ActiveSheet;
            sheet.get_Range("A1", Type.Missing).Value2 = "Process";
            sheet.get_Range("B1", Type.Missing).Value2 = "Memory";

            // Populate process list with 10 most expensive processes in terms of memory
            var res = (from proc in Process.GetProcesses()
                       orderby proc.WorkingSet64 descending
                       select new
                       {
                           Name = proc.ProcessName,
                           Memory = proc.WorkingSet64 / (1024 * 1024) /* in MB */
                       }
                      ).Take(10);

            int i = 2;
            foreach (var proc in res)
            {
                sheet.get_Range("A" + i, Type.Missing).Value2 = proc.Name;
                sheet.get_Range("B" + i, Type.Missing).Value2 = proc.Memory;
                i++;
            }

            // Add new chart
            Excel.Range range = sheet.get_Range("A1", Type.Missing);
            Excel.Chart chart = (Excel.Chart)excel.ActiveWorkbook.Charts.Add(
                Type.Missing,
                sheet,
                Type.Missing, Type.Missing);

            chart.ChartWizard(
                range.CurrentRegion, Type.Missing, Type.Missing, Type.Missing,
                Type.Missing, Type.Missing, Type.Missing,
                "Memory statistics",
                Type.Missing, Type.Missing, Type.Missing);

            chart.CopyPicture(
                Excel.XlPictureAppearance.xlScreen,
                Excel.XlCopyPictureFormat.xlBitmap,
                Excel.XlPictureAppearance.xlScreen);

            var word = new Word.Application();
            word.Visible = true;

            var missing = Type.Missing;
            word.Documents.Add(ref missing, ref missing, ref missing, ref missing);

            word.Selection.Paste();
        }
    }
}
