﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;

using Word = Microsoft.Office.Interop.Word;
using Excel = Microsoft.Office.Interop.Excel;

namespace Chapter22_OfficeInterop_NET40
{
    class Program
    {
        static void Main(string[] args)
        {
            var excel = new Excel.Application();
            excel.Visible = true;

            // New empty workbook
            excel.Workbooks.Add();

            // Column headers
            Excel.Worksheet sheet = excel.ActiveSheet;
            sheet.Range["A1"].Value = "Process";
            sheet.Range["B1"].Value = "Memory";

            // Populate process list with 10 most expensive processes in terms of memory
            var res = (from proc in Process.GetProcesses()
                       orderby proc.WorkingSet64 descending
                       select new
                       {
                           Name = proc.ProcessName,
                           Memory = proc.WorkingSet64 / (1024 * 1024) /* in MB */
                       }
                      ).Take(10);

            int i = 2;
            foreach (var proc in res)
            {
                sheet.Range["A" + i].Value = proc.Name;
                sheet.Range["B" + i].Value = proc.Memory;
                i++;
            }

            // Add new chart
            Excel.Range range = sheet.Range["A1"];
            Excel.Chart chart = excel.ActiveWorkbook.Charts.Add(After: sheet);

            chart.ChartWizard(range.CurrentRegion, Title: "Memory statistics");
            chart.CopyPicture();

            var word = new Word.Application();
            word.Visible = true;

            word.Documents.Add();
            word.Selection.Paste();
        }
    }
}
