﻿// Extern aliases for Chapter24_Assembly1 and Chapter24_Assembly2
extern alias X;
extern alias Y;

using Contoso.CashFlow.Client;

using System;
using System.Drawing;
using System.Windows;
using WpfPoint = System.Windows.Point;
using System.Linq;

class Program
{
    static void Main()
    {
        //
        // Imported namespace above. This is the same as writing:
        //   var t = new Contoso.CashFlow.Client.Terminal();
        //
        var t = new Terminal();

        //
        // Namespace clash, need to disambiguate the name of the type.
        //
        System.Drawing.Point p1 = new System.Drawing.Point { X = 5, Y = 3 };
        Console.WriteLine(p1.X);

        //
        // Or using an alias.
        //
        WpfPoint p2 = new WpfPoint { X = 5, Y = 3 };
        Console.WriteLine(p2.X);

        //
        // Using external aliases.
        //
        var fooFromX = new X::Bar.Foo();
        var fooFromY = new Y::Bar.Foo();

        //
        // Extension methods. Doesn't compile with System.Linq namespace imported.
        //
        var rabbits = new[] { 1, 1, 2, 3, 5, 8, 13, 21, 34, 55 };
        var evenRes = rabbits.Where(x => x % 2 == 0);
    }
}

// Shorthand declaration of namespace
namespace Contoso.CashFlow.Client
{
    public sealed class Terminal
    {

    }
}

// Alternative declaration of namespace
namespace Contoso
{
    namespace CashFlow
    {
        namespace Client
        {
            /* Already declared above
            public sealed class Terminal
            {

            }
             */
        }
    }
}

namespace System.Windows
{
    struct Point
    {
        public int X { get; set; }
        public int Y { get; set; }
    }
}

namespace System.Drawing
{
    struct Point
    {
        public int X { get; set; }
        public int Y { get; set; }
    }
}