﻿using System;
using System.ComponentModel;
using System.IO;
using System.Windows.Forms;

namespace Chapter29
{
    public partial class FileCopier : Form
    {
        public FileCopier()
        {
            InitializeComponent();
        }

        const int MB = 1024 * 1024;

        private void worker_DoWork(object sender, DoWorkEventArgs e)
        {
            var info = (CopyInfo)e.Argument;
            using (FileStream from = File.OpenRead(info.From))
            using (FileStream to = File.OpenWrite(info.To))
            {
                byte[] buffer = new byte[info.BlockSize];
                long total = from.Length, current = 0;
                while (!worker.CancellationPending)
                {
                    int n = from.Read(buffer, 0, buffer.Length);
                    to.Write(buffer, 0, n);

                    current += n;
                    worker.ReportProgress(
                        (int)(current / (total / 100)), // Trick to avoid overflow...
                        string.Format("{0}MB / {1}MB copied", current / MB, total / MB)
                    );

                    if (n < buffer.Length)
                        break;
                }

                if (worker.CancellationPending)
                {
                    e.Cancel = true;
                    // TODO: clean-up, e.g. deleting the target file.
                }
            }
        }

        private void worker_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            progressBar.Value = e.ProgressPercentage;
            lblProgress.Text = (string)e.UserState;
        }

        private void worker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            if (e.Cancelled)
                MessageBox.Show("File copy cancelled.", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else if (e.Error != null)
                MessageBox.Show(e.Error.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
            else
                MessageBox.Show("File copy completed.", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Information);

            lblFrom.Enabled = lblTo.Enabled = lblBlockSize.Enabled = txtFrom.Enabled = txtTo.Enabled = blockSize.Enabled = btnCopy.Enabled = btnCancel.Enabled = true;
        }

        private void btnCopy_Click(object sender, EventArgs e)
        {
            lblFrom.Enabled = lblTo.Enabled = lblBlockSize.Enabled = txtFrom.Enabled = txtTo.Enabled = blockSize.Enabled = btnCopy.Enabled = btnCancel.Enabled = false;

            worker.RunWorkerAsync(new CopyInfo
            {
                From = txtFrom.Text,
                To = txtTo.Text,
                BlockSize = (int)blockSize.Value
            });
        }
    }
}
