﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Chapter31
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void HungUI()
        {
            var frm = new HungForm();
            Application.Run(frm);
        }

        static void ClassicAPM()
        {
            // ReadFileAsync(/* put file name here */);
        }

        static void ReadFileAsync(string fileName)
        {
            FileStream fs = File.OpenRead(fileName);
            byte[] buffer = new byte[4 * 1024 * 1024]; // 4 MB
            fs.BeginRead(buffer, 0, buffer.Length, iar =>
            {
                int count = fs.EndRead(iar);
                // Process the buffer's elements in the range [0, count - 1].
            }, null);
        }

        static void ReadFileAsync_WithErrorHandling(string fileName, Form form, Action<byte[], int> onSuccess, Action<Exception> onError)
        {
            FileStream fs = File.OpenRead(fileName);
            byte[] buffer = new byte[4 * 1024 * 1024]; // 4 MB
            fs.BeginRead(buffer, 0, buffer.Length, iar =>
            {
                int count;
                try
                {
                    count = fs.EndRead(iar);
                }
                catch (IOException ex)
                {
                    // Handle the error.

                    // Post the operation to the UI message loop, e.g. in Windows Forms.
                    form.Invoke(new Action(() =>
                    {
                        onError(ex);
                    }));

                    return;
                }
                finally
                {
                    fs.Dispose();
                }

                // Post the operation to the UI message loop, e.g. in Windows Forms.
                form.Invoke(new Action(() =>
                {
                    // Process the buffer's elements in the range [0, count - 1]
                    // and render the result in the UI.
                    onSuccess(buffer, count);
                }));

            }, null);
        }

        static void EAP()
        {
            // DownloadStringAsyncWithEAP(/* put URI here */);
        }

        static void DownloadStringAsyncWithEAP(Uri uri)
        {
            WebClient client = new WebClient();
            client.DownloadStringCompleted += (o, e) =>
            {
                if (e.Error != null)
                {
                    try
                    {
                        throw e.Error; // Simply to restore structured exception handling?!
                    }
                    catch (WebException ex)
                    {
                        // Handle an error we can deal with.
                    }
                    return;
                }

                string html = e.Result;
                // Handle the result.
            };
            client.DownloadStringAsync(uri);
        }

        static void TAP()
        {
            // DownloadStringAsyncWithTAP(/* put URI here */);
        }

        static void DownloadStringAsyncWithTAP(Uri uri)
        {
            WebClient client = new WebClient();
            Task<string> html = client.DownloadStringTaskAsync(uri);
            // Handle result of the task, e.g. using ContinueWith or await.
        }

        static void AsyncMethods_UISynchronization()
        {
            new DownloadForm().ShowDialog();
        }

        static Task<T> ReturnAfter_Manual<T>(T value, int milliseconds, CancellationToken token)
        {
            var tcs = new TaskCompletionSource<T>();

            var ctr = default(CancellationTokenRegistration);

            var timer = new System.Threading.Timer(state =>
            {
                ((System.Threading.Timer)state).Dispose();
                ctr.Dispose(); // A struct, thus safe to dispose the default value.
                tcs.TrySetResult(value);
            });

            if (token.CanBeCanceled)
            {
                ctr = token.Register(() =>
                {
                    timer.Dispose();
                    tcs.TrySetCanceled();
                });
            }

            timer.Change(milliseconds, Timeout.Infinite /* no recurrence */);

            return tcs.Task;
        }

        static Task<T> ReturnAfter_ContinueWith<T>(T value, int milliseconds, CancellationToken token)
        {
            return Task.Delay(milliseconds, token)
                       .ContinueWith(t => value, TaskContinuationOptions.OnlyOnRanToCompletion);
        }

        static async Task<T> ReturnAfter_Await<T>(T value, int milliseconds, CancellationToken token)
        {
            await Task.Delay(milliseconds, token);
            return value;
        }

        static async Task<List<string>> DownloadSitesSequentialAsync(params string[] uris)
        {
            var result = new List<string>();

            using (var wc = new WebClient())
            {
                foreach (var uri in uris)
                    result.Add(await wc.DownloadStringTaskAsync(uri));
            }

            return result;
        }

        static async Task<List<string>> DownloadSitesParallelAsync(params string[] uris)
        {
            Task<string>[] sites = (from uri in uris
                                    let wc = new WebClient()
                                    select wc.DownloadStringTaskAsync(uri))
                                   .ToArray();

            string[] htmls = await Task.WhenAll(sites);
            return htmls.ToList();
        }

        static void CustomAwaiters()
        {
            new CustomAwaiterForm().ShowDialog();
        }
    }
}
